/*
 * Copyright (c) 2004, 2005, 2006, 2007 Luiz Otavio O Souza <loos.br@gmail.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

static const char rcsid[] = "$Id: sb.c,v 1.64 2007/07/28 18:08:01 luiz Exp $";

#include <sys/types.h>
#include <sys/socket.h>

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "sb.h"
#include "sql.h"
#include "net-io.h"
#include "return.h"
#include "protocol.h"
#include "check-cmd.h"
#include "msn-proxy.h"

void
sb_free_xfr(sb_ *sb) {
    if (sb->xfr) {
	str_free(sb->xfr);
	free(sb->xfr);
	sb->xfr = (string *)0;
    }
}

sb_ *
sb_alloc(user_ *user) {
 sb_		*sb;

    sb = (sb_ *)malloc(sizeof(sb_));
    if (sb == (sb_ *)0)
	die_nomem();
    bzero(sb, sizeof(sb_));

    sb->listen.ev_fd = -1;
    sb->user	     = user;
    sb->closed	     = NO;
    sb->warned	     = NO;

    /* create a new sb entry on sql */
    sb->id = sql_insert_sb(user);
    if (sb->id <= 0) {
	free (sb);
	return((sb_ *)0);
    }

    /* initialize user list */
    SLIST_INIT(&sb->user_head);

    /* insert this new sb on user sb_list */
    LIST_INSERT_HEAD(&user->sb_head, sb, entries);

    return(sb);
}

struct sb_user_entry *
sb_user_entry_alloc(void) {
 struct sb_user_entry	*entry;

    entry = (struct sb_user_entry *)malloc(sizeof(struct sb_user_entry));
    if (entry == (struct sb_user_entry *)0)
	die_nomem();

    bzero(entry, sizeof(struct sb_user_entry));
    return(entry);
}

void
sb_connect(const int listenfd, short event, void *p) {
 struct	sockaddr	client_sa;
 socklen_t		client_sa_len;
 log_			*log = &config.log;
 int			client_fd;
 sb_			*sb = p;

    if ((event & EV_TIMEOUT)) {
	log->debug("debug: sb proxy connection timeout\n");
	sb = sb_disconnect(sb);
	goto end;
    }

    /* prepare client sockaddr */
    client_sa_len = sizeof(struct sockaddr);
    bzero(&client_sa, client_sa_len);

    /* accept connection */
    client_fd = accept(listenfd, &client_sa, &client_sa_len);
    if (client_fd < 0) {
	log->debug("debug: unable to accept new sb client connection\n");
	sb = sb_disconnect(sb);
	goto end;
    }

    /* set socket options */
    if (set_options(client_fd) == -1 || set_nonblock(client_fd) == -1) {
	log->debug("debug: unable to set socket options on sb client fd\n");
	sb = sb_disconnect(sb);
	goto end;
    }

    /* attach server */
    sb->server = server_connect2(sb->xfr, sb_server_read, sb_server_write, sb);
    if (sb->server == (server_ *)0) {
	log->debug("debug: fail to connect to sb server\n");
	sb = sb_disconnect(sb);
	goto end;
    }

    /* free xfr */
    sb_free_xfr(sb);

    /* attach client */
    sb->client = client_alloc(client_fd, sb_client_read, sb_client_write, sb);

    /* reset queue events */
    if (EVENT_FD((&sb->listen)) != -1) {
	event_del(&sb->listen);
	sb->listen.ev_fd = -1;
    }

end:
    /* close listenfd */
    while (close(listenfd) != 0 && errno == EINTR);
}

sb_ *
sb_disconnect(sb_ *sb) {
 struct sb_user_entry	*entry;
 user_			*user = sb->user;

    /* log stop */
    if (sb->closed == NO)
	sql_log_stop(sb, &user->email, &user->dn);

    /* remove sb session from sql */
    sql_remove_sb(sb->id);

    /* close connection to NS */
    sb->server = server_disconnect(sb->server);

    /* close connection to NS client */
    sb->client = client_disconnect(sb->client);

    /* reset queue events */
    if (EVENT_FD((&sb->listen)) != -1) {
	event_del(&sb->listen);
	sb->listen.ev_fd = -1;
    }

    /* free xfr if exist */
    sb_free_xfr(sb);

    /* free start if exist */
    str_free(&sb->start);

    /* free sb->users */
    while (!SLIST_EMPTY(&sb->user_head)) {
	entry = SLIST_FIRST(&sb->user_head);
	SLIST_REMOVE_HEAD(&sb->user_head, entries);
	free(entry);
    }

    LIST_REMOVE(sb, entries);

    free(sb);

    return((sb_ *)0);
}

void
sb_disconnect_all(user_ *user) {
 sb_		*sb;

    while (!LIST_EMPTY(&user->sb_head)) {
	sb = LIST_FIRST(&user->sb_head);
	sb_disconnect(sb);
    }
}

void
sb_client_write(const int evfd, short event, void *p) {
 sb_			*sb       = p;
 log_			*log      = &config.log;
 user_			*user     = sb->user;
 client_		*client   = sb->client;
 server_		*server   = sb->server;
 client_		*nsclient = user->ns.client;
 command		*cmd      = server->commands.cmd;
 const char		*to       = "sb client";

    (void) evfd;

    if ((event & EV_TIMEOUT)) {
	log->debug("debug: sb client write timeout\n");
	sb = sb_disconnect(sb);
	return;
    }

    if (!HAS_CMD(server))
	return;

    /* hook for send commands */
    if (check_sb_cmd(sb, cmd, SERVER_SB_CMD) == RFAIL) {

	log->debug("debug: sb server command not accepted\n");
	sb = sb_disconnect(sb);
	return;
    }

    /* flush the server command queue */
    if (send_command(client->fd, cmd, &user->email, to) == RFAIL) {

	log->debug("debug: fail to send sb client command\n");
	sb = sb_disconnect(sb);
	return;
    }

    /* hook for post send commands */
    switch (check_sb_cmd(sb, cmd, SERVER_SB_POST_CMD)) {

	case RFAIL:
	    log->debug("debug: fail to run sb server post command\n");
	    /* fallthrough */

	case DISCONNECT:
	    sb = sb_disconnect(sb);
	    /* fallthrough */

	case RETURN:
	    return;
    }

    /* command >> 1 */
    shift_server_commands(server, cmd);
    free_command(cmd);

    client_sched_read(client);
    client_sched_read(nsclient);

    if (!HAS_CMD(server))
	return;

    client_sched_write(client);
}

void sb_client_read(const int evfd, short event, void *p) {
 sb_			*sb	  = p;
 log_			*log	  = &config.log;
 client_		*client   = sb->client;
 server_		*server   = sb->server;
 client_		*nsclient = sb->user->ns.client;

    (void) evfd;

    if ((event & EV_TIMEOUT)) {
	log->debug("debug: sb client read timeout\n");
	sb = sb_disconnect(sb);
	return;
    }

    /* read user command */
    switch (read_client_command(client)) {

	case AGAIN:
	    client_sched_read(client);
	    client_sched_read(nsclient);
	    return;

	case END:
	    log->debug("debug: connection closed by sb client\n");
	    sb = sb_disconnect(sb);
	    return;

	case RFAIL:
	    log->debug("debug: fail to read sb client command\n");
	    sb = sb_disconnect(sb);
	    return;
    }

    if (HAS_CMD(client))
	server_sched_write(server);

    client_sched_read(client);
    client_sched_read(nsclient);
}

void
sb_server_read(int evfd, short event, void *p) {
 sb_			*sb	  = p;
 log_			*log	  = &config.log;
 client_		*client   = sb->client;
 server_		*server   = sb->server;
 server_		*nsserver = sb->user->ns.server;

    (void) evfd;

    if ((event & EV_TIMEOUT)) {
	log->debug("debug: sb server read timeout\n");
	sb = sb_disconnect(sb);
        return;
    }

    switch (read_server_command(server)) {

	case AGAIN:
	    server_sched_read(server);
	    server_sched_read(nsserver);
	    return;

	case END:
            /* connection reset by peer */
            
	    if (HAS_CMD(server)) {

		/*
		 * ns server queue is not empty so close server connection
		 * and schedule another client write
		 */
		server_close(server);

		client_sched_write(client);

		log->debug("debug: closed connection to ns server. flushing queue.\n");
		return;
            }

	    log->debug("debug: connection closed by sb server\n");
	    sb = sb_disconnect(sb);
	    return;

	case RFAIL:
	    log->debug("debug: fail to read sb server command\n");
	    sb = sb_disconnect(sb);
	    return;
    }

    if (HAS_CMD(server))
        client_sched_write(client);

    server_sched_read(server);
    server_sched_read(nsserver);
}

void
sb_server_write(int evfd, short event, void *p) {
 sb_			*sb	= p;
 log_			*log	= &config.log;
 user_			*user	= sb->user;
 client_		*client = sb->client;
 server_		*server = sb->server;
 command		*cmd	= client->commands.cmd;
 const char		*to	= "sb server";

    (void) evfd;

    if ((event & EV_TIMEOUT)) {
	log->debug("debug: sb server write timeout\n");
	sb = sb_disconnect(sb);
	return;
    }

    if (!HAS_CMD(client))
	return;

    /* hook for send commands */
    if (check_sb_cmd(sb, cmd, CLIENT_SB_CMD) == RFAIL) {
	log->debug("debug: client sb command not accepted\n");
	sb = sb_disconnect(sb);
	return;
    }

    if (send_command(server->fd, cmd, &user->email, to) == RFAIL) {
	log->debug("debug: fail to send sb server command\n");
	sb = sb_disconnect(sb);
	return;
    }

    /* hook for post send commands */
    switch (check_sb_cmd(sb, cmd, CLIENT_SB_POST_CMD)) {

	case RFAIL:
	    log->debug("debug: client sb post command not accepted\n");
	    /* fallthrough */

	case DISCONNECT:
	    sb = sb_disconnect(sb);
	    /* fallthrough */

	case RETURN:
	    return;
    }

    /* command >> 1 */
    shift_client_commands(client, cmd);
    free_command(cmd);

    if (!HAS_CMD(client))
	return;

    server_sched_write(server);
}

sb_ *
sb_proxy(user_ *user, string **arg) {
 sb_                    *sb;
 int                    listenfd;
 string                 *port;

    /* alloc new sb */
    sb = sb_alloc(user);
    if (sb == (sb_ *)0)
	return((sb_ *)0);

    port = bind_free_port(config.port_min, config.port_max,
			&config.defaults.internal_host, &listenfd);
    if (port == (string *)0) {

	/* cannot allocate any port */
	sb = sb_disconnect(sb);
	return((sb_ *)0);
    }

    /* save old xfr host */
    sb->xfr = *arg;

    /* setup new xfr host */
    *arg = port;

    /* event set */
    event_set(&sb->listen, listenfd, EV_READ, sb_connect, sb);
    event_add(&sb->listen, &config.timeout_listen);

    return(sb);
}

int
xfr_sb_proxy(user_ *user, command *cmd, int args) {
 sb_			*sb;
 string			**address;

    if (cmd->args_len < 3)
	return(RFAIL);

    address = &cmd->args[2];

    /* proxy sb connection */
    sb = sb_proxy(user, address);
    if (sb == (sb_ *)0)
	return(RFAIL);

    /* save start email */
    if (msn_decode(&user->email, &sb->start) == RFAIL)
	return(RFAIL);

    if (sql_log_start(sb->id, &user->email, &user->dn) == RFAIL)
	return(RFAIL);

    return(ROK);
}

int
rng_sb_proxy(user_ *user, command *cmd, int args) {
 sb_			*sb;
 string			dn;
 string			*dn_   = cmd->args[5];
 string			*email = cmd->args[4];
 string			**address = &cmd->args[1];

    /* user is allowed to chat with us ? */
    switch (sql_check_contact_allowed(&user->email, email)) {
	case RFAIL:
	    return(RFAIL);
	case NO:
	    return(RETURN);
	case YES:
	default:
	    break;
    }

    /* proxy sb connection */
    sb = sb_proxy(user, address);
    if (sb == (sb_ *)0)
	return(RFAIL);

    /* save start email */
    if (msn_decode(email, &sb->start) == RFAIL)
	return(RFAIL);

    str_zero(&dn);
    if (msn_decode(dn_, &dn) == RFAIL)
        return(RFAIL);

    if (sql_log_start(sb->id, email, &dn) == RFAIL) {
	str_free(&dn);
	return(RFAIL);
    }

    str_free(&dn);

    if (sql_log_join(sb->id, &user->email, &user->dn) == RFAIL)
	return(RFAIL);

    return(ROK);
}

int
msn_sb_joi(sb_ *sb, command *cmd, int args) {
 struct sb_user_entry	*entry;
 string			*dn_   = cmd->args[1];
 string			*email = cmd->args[0];
 string			dn;

    if (cmd->args_len < 2)
	return(RFAIL);

    /*
     * save user email - this may be a multiuser chat, so this information
     * is needed when we receive a BYE command from user
     */
    entry = sb_user_entry_alloc();
    if (str_copy(&entry->email, email->s, email->len) == 0)
	die_nomem();
    SLIST_INSERT_HEAD(&sb->user_head, entry, entries);

    /* decode display name */
    str_zero(&dn);
    if (msn_decode(dn_, &dn) == RFAIL)
	return(RFAIL);

    if (sql_log_join(sb->id, email, &dn) == RFAIL) {
	str_free(&dn);
	return(RFAIL);
    }

    str_free(&dn);
    return(ROK);
}

int
msn_sb_post_out(sb_ *sb, command *cmd, int args) {
    return(DISCONNECT);
}

int
sb_iro(sb_ *sb, command *cmd, int args) {
 struct sb_user_entry	*entry;
 string			*email;
 string			*dn_;
 string			dn;
 log_			*log = &config.log;

    str_zero(&dn);

    if (cmd->args_len < 5) {
	log->debug("debug: ARGS_LEN != IRO_LEN\n");
	print_command(cmd);
        return(RFAIL);
    }

    /* to access any pointer inside a command, check args_len first */
    dn_	  = cmd->args[4];
    email = cmd->args[3];

    /*
     * save user email - this may be a multiuser chat, so this information
     * is needed when we receive a BYE command from user
     */
    entry = sb_user_entry_alloc();
    if (str_copy(&entry->email, email->s, email->len) == 0)
	die_nomem();
    SLIST_INSERT_HEAD(&sb->user_head, entry, entries);

    if (msn_decode(dn_, &dn) == RFAIL) {

	log->debug("debug: fail to decode SB_IRO\n");
	str_free(&dn);
	return(RFAIL);
    }

    if (sb->start.len > 0 && email->len > 0 &&
	    strcmp(sb->start.s, email->s) != 0) {

	/* principal has not started the connection - log join */
	if (sql_log_join(sb->id, email, &dn) == RFAIL) {
	    str_free(&dn);
	    return(RFAIL);
	}
    }

    str_free(&dn);
    return(ROK);
}

int
msn_sb_post_bye(sb_ *sb, command *cmd, int args) {
 struct sb_user_entry	*entry;
 user_			*user = sb->user;

    if (cmd->args_len < 1)
	return(RFAIL);

    SLIST_FOREACH(entry, &sb->user_head, entries) {

	if (strcmp(entry->email.s, cmd->args[0]->s) == 0) {

	    SLIST_REMOVE(&sb->user_head, entry, sb_user_entry, entries);

	    /* log stop */
	    if (strcmp(user->email.s, entry->email.s) == 0)
		sql_log_stop(sb, &user->email, &user->dn);
	    else
		sql_log_stop2(sb, &user->email, &entry->email);

	    break;
	}
    }

    /* last user on this sb connection */
    if (SLIST_EMPTY(&sb->user_head)) {
	sb->closed = YES;
	return(DISCONNECT);
    }

    return(ROK);
}

int
msn_sb_cal(sb_ *sb, command *cmd, int args) {
 user_			*user = sb->user;
 string                 *email = cmd->args[1];

    if (cmd->args_len < 2)
	return(RFAIL);

    /* user is allowed to chat with us ? */
    switch (sql_check_contact_allowed(&user->email, email)) {
	case RFAIL:
	    return(RFAIL);
	case NO:
	    return(DISCONNECT);
	case YES:
	default:
	    break;
    }
    return(ROK);
}

